#include "collision_manager_alg_node.h"

CollisionManagerAlgNode::CollisionManagerAlgNode(void) :
  algorithm_base::IriBaseAlgorithm<CollisionManagerAlgorithm>()
  ,tf2_listener(tf2_buffer)
{
  //init class attributes if necessary
  if(!this->private_node_handle_.getParam("rate", this->config_.rate))
  {
    ROS_WARN("CollisionManagerAlgNode::CollisionManagerAlgNode: param 'rate' not found");
  }
  else
    this->setRate(this->config_.rate);

  if(!this->private_node_handle_.getParam("fixed_frame", this->config_.fixed_frame))
  {
    ROS_WARN("AdcLandmarksSlamSolverAlgNode::AdcLandmarksSlamSolverAlgNode: param 'fixed_frame' not found. Setting it to map.");
    this->config_.fixed_frame = "map";
  }

  this->new_imu_input_ = false;
  this->imu_frame_ = "";
  this->tf_loaded_ = false;
  this->in_collision_ = false;
  this->low_acc_count_ = 0;
  this->is_front_icp_input_data_ok_ = false;
  this->is_rear_icp_input_data_ok_ = false;
  // [init publishers]
  
  // [init subscribers]
  this->rear_laser_scan_subscriber_ = this->private_node_handle_.subscribe("rear_laser_scan", 1, &CollisionManagerAlgNode::rear_laser_scan_callback, this);
  pthread_mutex_init(&this->rear_laser_scan_mutex_,NULL);

  this->front_laser_scan_subscriber_ = this->private_node_handle_.subscribe("front_laser_scan", 1, &CollisionManagerAlgNode::front_laser_scan_callback, this);
  pthread_mutex_init(&this->front_laser_scan_mutex_,NULL);

  this->imu_subscriber_ = this->private_node_handle_.subscribe("imu", 1, &CollisionManagerAlgNode::imu_callback, this);
  pthread_mutex_init(&this->imu_mutex_,NULL);

  
  // [init services]
  
  // [init clients]
  icp_client_ = this->private_node_handle_.serviceClient<iri_laser_scan_icp::icp>("icp");

  
  // [init action servers]
  
  // [init action clients]
}

CollisionManagerAlgNode::~CollisionManagerAlgNode(void)
{
  // [free dynamic memory]
  pthread_mutex_destroy(&this->rear_laser_scan_mutex_);
  pthread_mutex_destroy(&this->front_laser_scan_mutex_);
  pthread_mutex_destroy(&this->imu_mutex_);
}

void CollisionManagerAlgNode::mainNodeThread(void)
{
  //lock access to algorithm if necessary
  // this->alg_.lock();
  // ROS_DEBUG("CollisionManagerAlgNode::mainNodeThread");
  this->imu_mutex_enter();
  this->rear_laser_scan_mutex_enter();
  this->front_laser_scan_mutex_enter();

  if (this->new_imu_input_)
  {
    Eigen::Vector3d global_acc, global_ang_vel;
    if (get_global_imu_data(global_acc, global_ang_vel))//Transform data to fixed frame coordinates
    {
      double acc_norm_2d = std::sqrt(std::pow(global_acc(0),2) + std::pow(global_acc(1),2));
      if (this->in_collision_)
      {
        double max_acc_norm_2d = std::sqrt(std::pow(this->max_acc_(0),2) + std::pow(this->max_acc_(1),2));
        if (check_end_of_collision(acc_norm_2d))
        {
          double collision_angle = std::atan2(this->max_acc_(1), this->max_acc_(0));
          while (collision_angle >= M_PI)
            collision_angle -= 2*M_PI;
          while (collision_angle < -M_PI)
            collision_angle += 2*M_PI;
          ROS_INFO_STREAM("CollisionManagerAlgNode::mainNodeThread -> collision end detected.");
          ROS_INFO_STREAM("CollisionManagerAlgNode::mainNodeThread -> collision detected: acc = " << max_acc_norm_2d << "; angle = " << collision_angle);

          //update rot_angle_estimated
          this->front_icp_srv_.request.rot_angle_estimation += global_ang_vel(2)*(this->imu_stamp_ - this->last_imu_t_).toSec();
          this->rear_icp_srv_.request.rot_angle_estimation = -this->front_icp_srv_.request.rot_angle_estimation;

          // Call the service
          if (!(this->config_.front_laser_en || this->config_.rear_laser_en))
            ROS_WARN("CollisionManagerAlgNode::mainNodeThread -> Collision end detected but no laser is enabled.");
          else
          {
            if (this->is_front_icp_input_data_ok_ && set_last_ranges(true))
            {
              if (icp_client_.call(front_icp_srv_))
              {
               ROS_INFO("CollisionManagerAlgNode::mainNodeThread -> Front icp_client_ received a response from service server");
               ROS_INFO_STREAM("   valid " << (int)front_icp_srv_.response.valid);
               ROS_INFO_STREAM("   new_laser_pose " << front_icp_srv_.response.new_laser_pose);
               ROS_INFO_STREAM("   covariance " << front_icp_srv_.response.covariance[0] << ", " << front_icp_srv_.response.covariance[1] << ", " << front_icp_srv_.response.covariance[2] << ", " << front_icp_srv_.response.covariance[3] << ", " << front_icp_srv_.response.covariance[4] << ", " << front_icp_srv_.response.covariance[5] << ", " << front_icp_srv_.response.covariance[6] << ", " << front_icp_srv_.response.covariance[7] << ", " << front_icp_srv_.response.covariance[8]);
               ROS_INFO_STREAM("   nvalid " << front_icp_srv_.response.nvalid);
               ROS_INFO_STREAM("   error " << front_icp_srv_.response.error);
              }
              else
               ROS_ERROR("CollisionManagerAlgNode::mainNodeThread -> Failed to call service on topic %s",this->icp_client_.getService().c_str());
            }
            if (this->is_rear_icp_input_data_ok_ && set_last_ranges(false))
            {
              if (icp_client_.call(rear_icp_srv_))
              {
               ROS_INFO("CollisionManagerAlgNode::mainNodeThread -> Rear icp_client_ received a response from service server");
               ROS_INFO_STREAM("   valid " << (int)rear_icp_srv_.response.valid);
               ROS_INFO_STREAM("   new_laser_pose " << rear_icp_srv_.response.new_laser_pose);
               ROS_INFO_STREAM("   covariance " << rear_icp_srv_.response.covariance[0] << ", " << rear_icp_srv_.response.covariance[1] << ", " << rear_icp_srv_.response.covariance[2] << ", " << rear_icp_srv_.response.covariance[3] << ", " << rear_icp_srv_.response.covariance[4] << ", " << rear_icp_srv_.response.covariance[5] << ", " << rear_icp_srv_.response.covariance[6] << ", " << rear_icp_srv_.response.covariance[7] << ", " << rear_icp_srv_.response.covariance[8]);
               ROS_INFO_STREAM("   nvalid " << rear_icp_srv_.response.nvalid);
               ROS_INFO_STREAM("   error " << rear_icp_srv_.response.error);
              }
              else
               ROS_ERROR("CollisionManagerAlgNode::mainNodeThread -> Failed to call service on topic %s",this->icp_client_.getService().c_str());
            }
          }
          //restart variables
          this->in_collision_ = false;
          this->low_acc_count_ = 0;
          this->is_front_icp_input_data_ok_ = false;
          this->is_rear_icp_input_data_ok_ = false;
        }
        else 
        {
          //Update max_acc
          if (acc_norm_2d > max_acc_norm_2d)
            this->max_acc_ = global_acc;

          //Update angle rotated
          this->front_icp_srv_.request.rot_angle_estimation += global_ang_vel(2)*(this->imu_stamp_ - this->last_imu_t_).toSec();
          this->last_imu_t_ = this->imu_stamp_;

          //Check collision timeout
          if ((ros::Time::now() - this->collision_start_stamp_) > ros::Duration(this->config_.collision_timeout))
          {
            ROS_WARN_STREAM("CollisionManagerAlgNode::mainNodeThread -> No end of collision detected after " << this->config_.collision_timeout << " seconds.");
            this->in_collision_ = false;
            this->low_acc_count_ = 0;
          }
        }
      }
      else if (check_start_of_collision(acc_norm_2d))
      {
        ROS_INFO_STREAM("CollisionManagerAlgNode::mainNodeThread -> collision start detected.");
        this->collision_start_stamp_ = this->imu_stamp_;

        // Get reference scan
        if (!(this->config_.front_laser_en || this->config_.rear_laser_en))
          ROS_WARN("CollisionManagerAlgNode::mainNodeThread -> Collision detected but no laser is enabled.");
        else
        {
          this->is_front_icp_input_data_ok_ = set_ref_ranges(this->collision_start_stamp_, true);
          this->is_rear_icp_input_data_ok_ = set_ref_ranges(this->collision_start_stamp_, false);
        }

        //init variables
        this->in_collision_ = true;
        this->max_acc_ = global_acc;
        this->last_imu_t_ = this->collision_start_stamp_;
        this->front_icp_srv_.request.rot_angle_estimation = 0.0;
        this->rear_icp_srv_.request.rot_angle_estimation = 0.0;
      }
    }
    this->new_imu_input_ = false;
  }

  // [fill msg structures]
  
  /*
  //tf2_listener example BEGIN
  try{
    std::string target_frame             = "child_frame";
    std::string source_frame             = "parent_frame";
    ros::Time time                       = ros::Time::now();
    ros::Duration timeout                = ros::Duration(0.1);
    this->alg_.unlock();
    bool tf2_exists = this->tf2_buffer.canTransform(target_frame, source_frame, time, timeout);
    this->alg_.lock();
    if(tf2_exists){
      geometry_msgs::PoseStamped stamped_pose_in;
      stamped_pose_in.header.stamp     = time;
      stamped_pose_in.header.frame_id  = source_frame;
      stamped_pose_in.pose.position.x    = 1.0;
      stamped_pose_in.pose.position.y    = 0.0;
      stamped_pose_in.pose.position.z    = 0.0;
      tf2::Quaternion quat_tf;
      quat_tf.setRPY(0.0, 0.0, 1.5709);
      geometry_msgs::Quaternion quat_msg;
      tf2::convert(quat_tf, quat_msg);
      stamped_pose_in.pose.orientation = quat_msg;
      double yaw, pitch, roll;
      tf2::Matrix3x3(tf2::Quaternion(quat_msg.x, quat_msg.y,quat_msg.z, quat_msg.w)).getEulerYPR(yaw, pitch, roll);
      ROS_INFO("Original    pose in '%s' frame, with (x,y,z)=[%f,%f,%f], (yaw,pitch,roll)=[%f,%f,%f]", stamped_pose_in.header.frame_id.c_str(), stamped_pose_in.pose.position.x, stamped_pose_in.pose.position.y, stamped_pose_in.pose.position.z, yaw, pitch, roll);
      geometry_msgs::PoseStamped stamped_pose_out;
      geometry_msgs::TransformStamped transform;
      transform = this->tf2_buffer.lookupTransform(target_frame, source_frame, time);
      quat_msg = transform.transform.rotation;
      tf2::Matrix3x3(tf2::Quaternion(quat_msg.x, quat_msg.y,quat_msg.z, quat_msg.w)).getEulerYPR(yaw, pitch, roll);
      ROS_INFO("Found transform betwen frames (%s-->%s) with (x,y,z)=[%f,%f,%f], (yaw,pitch,roll)=[%f,%f,%f]", source_frame.c_str(), target_frame.c_str(), transform.transform.translation.x, transform.transform.translation.y, transform.transform.translation.z, yaw, pitch, roll);
      tf2::doTransform(stamped_pose_in, stamped_pose_out, transform);
      quat_msg = stamped_pose_out.pose.orientation;
      tf2::Matrix3x3(tf2::Quaternion(quat_msg.x, quat_msg.y,quat_msg.z, quat_msg.w)).getEulerYPR(yaw, pitch, roll);
      ROS_INFO("Transformed pose in '%s' frame, with (x,y,z)=[%f,%f,%f], (yaw,pitch,roll)=[%f,%f,%f]", stamped_pose_out.header.frame_id.c_str(), stamped_pose_out.pose.position.x, stamped_pose_out.pose.position.y, stamped_pose_out.pose.position.z, yaw, pitch, roll);
      ROS_INFO("---");
    }else{
      ROS_WARN("No transform found from '%s' to '%s'", source_frame.c_str(), target_frame.c_str()); }
  }catch (tf2::TransformException &ex){
    ROS_ERROR("TF2 Exception: %s",ex.what()); }
  //tf2_listener example END
  */

  
  // [fill srv structure and make request to the server]
  //icp_srv_.request.data = my_var;
  //ROS_INFO("CollisionManagerAlgNode:: Calling service icp_client_!");
  //if (icp_client_.call(icp_srv_))
  //{
  //  ROS_INFO("CollisionManagerAlgNode:: icp_client_ received a response from service server");
  ////  ROS_INFO("CollisionManagerAlgNode:: Response: %s", icp_srv_.response.somestringfield.c_str());
  //}
  //else
  //{
  //  ROS_INFO("CollisionManagerAlgNode:: Failed to call service on topic %s",this->icp_client_.getService().c_str());
  //}


  
  // [fill action structure and make request to the action server]

  // [publish messages]

  this->front_laser_scan_mutex_exit();
  this->rear_laser_scan_mutex_exit();
  this->imu_mutex_exit();
  // this->alg_.unlock();
}

bool CollisionManagerAlgNode::get_global_imu_data(Eigen::Vector3d& _global_acc, Eigen::Vector3d& _global_ang_vel)
{
  if (this->config_.fixed_frame != this->imu_frame_ || !this->tf_loaded_)
  {
    geometry_msgs::TransformStamped tf_fixed_frame_imu_msg;
    if (!this->tf2_buffer.canTransform(this->config_.fixed_frame, this->imu_frame_, ros::Time::now(), ros::Duration(this->config_.tf_timeout)))
    {
      ROS_WARN_STREAM("CollisionManagerAlgNode::get_global_imu_data -> Can't transform from " << this->config_.fixed_frame << " to " << this->imu_frame_ << " at " << ros::Time::now());
      this->tf_loaded_ = false;
      return false;
    }
    tf_fixed_frame_imu_msg = this->tf2_buffer.lookupTransform(this->config_.fixed_frame, this->imu_frame_, ros::Time::now());
    Eigen::Quaternion<double> rot(
    tf_fixed_frame_imu_msg.transform.rotation.w, tf_fixed_frame_imu_msg.transform.rotation.x, tf_fixed_frame_imu_msg.transform.rotation.y, tf_fixed_frame_imu_msg.transform.rotation.z);
    
    this->tf_fixed_frame_imu_ = Eigen::Transform<double,3,Eigen::Affine>(rot);
    this->tf_loaded_ = true;
  }
  _global_acc = this->tf_fixed_frame_imu_*this->imu_local_acc_;
  _global_ang_vel = this->tf_fixed_frame_imu_*this->imu_local_ang_vel_;
  return true;
}

bool CollisionManagerAlgNode::check_collision_transition(double _acc_norm_2d, bool _start)
{
  if (this->config_.collision_transition_counter_en)
  {
    if ((_start && (_acc_norm_2d > this->config_.collision_acc_th)) || (!_start && (_acc_norm_2d < this->config_.collision_acc_th)))
      this->low_acc_count_++;
    else
    {
      this->low_acc_count_ = 0;
      return false;
    }
  }
  return ((_start && (_acc_norm_2d > this->config_.collision_acc_th)) || (!_start && (_acc_norm_2d < this->config_.collision_acc_th)))  
         && (_start || (!_start && (!this->config_.collision_transition_counter_en || (this->low_acc_count_ >= this->config_.collision_counter_limit))));
}

bool CollisionManagerAlgNode::check_end_of_collision(double _acc_norm_2d)
{
  return check_collision_transition(_acc_norm_2d, false);
}

bool CollisionManagerAlgNode::check_start_of_collision(double _acc_norm_2d)
{
  return check_collision_transition(_acc_norm_2d, true);
}

bool CollisionManagerAlgNode::set_ref_ranges(ros::Time _stamp, bool _front)
{
  if (!(_front? this->config_.front_laser_en: this->config_.rear_laser_en))
    return false;
  auto it = (_front? this->front_ranges_.begin(): this->rear_ranges_.begin());
  while ((it != (_front? this->front_ranges_.end(): this->rear_ranges_.end())) && (it->stamp > (this->collision_start_stamp_ - ros::Duration(this->config_.scan_diff_t_from_collision))))
    ++it;
  if (it == (_front? this->front_ranges_.end(): this->rear_ranges_.end()))
  {
    ROS_WARN_STREAM("CollisionManagerAlgNode::set_ref_ranges -> Collision detected at " << this->collision_start_stamp_ << " but there is no " 
                      << (_front? "front ": "rear ") << "scan before collision.");
    if ((_front? this->front_ranges_: this->rear_ranges_).size() > 0)
      ROS_WARN_STREAM("CollisionManagerAlgNode::set_ref_ranges -> scan times: " << (_front? this->front_ranges_: this->rear_ranges_).front().stamp << ", " << (_front? this->front_ranges_: this->rear_ranges_).back().stamp);
    else
      ROS_WARN_STREAM("CollisionManagerAlgNode::set_ref_ranges -> No scan buffered");
    return false;
  }
  (_front? this->front_icp_srv_: this->rear_icp_srv_).request.ref_ranges = it->ranges;
  return true;
}

bool CollisionManagerAlgNode::set_last_ranges(bool _front)
{
  if (!(_front? this->config_.front_laser_en: this->config_.rear_laser_en))
    return false;

  if ((_front? this->front_ranges_: this->rear_ranges_).size() > 0)
  {
    (_front? this->front_icp_srv_: this->rear_icp_srv_).request.last_ranges = (_front? this->front_ranges_: this->rear_ranges_).front().ranges;
    return true;
  }
  ROS_WARN_STREAM("CollisionManagerAlgNode::set_last_ranges -> No scan buffered");
  return false;
}

/*  [subscriber callbacks] */
void CollisionManagerAlgNode::rear_laser_scan_callback(const sensor_msgs::LaserScan::ConstPtr& msg)
{
  // ROS_INFO("CollisionManagerAlgNode::rear_laser_scan_callback: New Message Received");

  //use appropiate mutex to shared variables if necessary
  //this->alg_.lock();
  this->rear_laser_scan_mutex_enter();
  if (this->config_.rear_laser_en)
  {
    rear_icp_srv_.request.angle_min = msg->angle_min;
    rear_icp_srv_.request.angle_max = msg->angle_max;
    rear_icp_srv_.request.angle_increment = msg->angle_increment;
    rear_icp_srv_.request.scan_time = msg->scan_time;
    rear_icp_srv_.request.range_min = msg->range_min;
    rear_icp_srv_.request.range_max = msg->range_max;

    RangesWithStamp r;
    r.ranges = msg->ranges;
    r.stamp = msg->header.stamp;
    if (this->rear_ranges_.size() >= this->config_.scan_buffer_size)
      this->rear_ranges_.pop_back();
    this->rear_ranges_.push_front(r);
  }
  //std::cout << msg->data << std::endl;
  //unlock previously blocked shared variables
  //this->alg_.unlock();
  this->rear_laser_scan_mutex_exit();
}

void CollisionManagerAlgNode::rear_laser_scan_mutex_enter(void)
{
  pthread_mutex_lock(&this->rear_laser_scan_mutex_);
}

void CollisionManagerAlgNode::rear_laser_scan_mutex_exit(void)
{
  pthread_mutex_unlock(&this->rear_laser_scan_mutex_);
}

void CollisionManagerAlgNode::front_laser_scan_callback(const sensor_msgs::LaserScan::ConstPtr& msg)
{
  // ROS_INFO("CollisionManagerAlgNode::front_laser_scan_callback: New Message Received");

  //use appropiate mutex to shared variables if necessary
  //this->alg_.lock();
  this->front_laser_scan_mutex_enter();
  if (this->config_.front_laser_en)
  {
    front_icp_srv_.request.angle_min = msg->angle_min;
    front_icp_srv_.request.angle_max = msg->angle_max;
    front_icp_srv_.request.angle_increment = msg->angle_increment;
    front_icp_srv_.request.scan_time = msg->scan_time;
    front_icp_srv_.request.range_min = msg->range_min;
    front_icp_srv_.request.range_max = msg->range_max;

    RangesWithStamp r;
    r.ranges = msg->ranges;
    r.stamp = msg->header.stamp;
    if (this->front_ranges_.size() >= this->config_.scan_buffer_size)
      this->front_ranges_.pop_back();
    this->front_ranges_.push_front(r);
  }

  //std::cout << msg->data << std::endl;
  //unlock previously blocked shared variables
  //this->alg_.unlock();
  this->front_laser_scan_mutex_exit();
}

void CollisionManagerAlgNode::front_laser_scan_mutex_enter(void)
{
  pthread_mutex_lock(&this->front_laser_scan_mutex_);
}

void CollisionManagerAlgNode::front_laser_scan_mutex_exit(void)
{
  pthread_mutex_unlock(&this->front_laser_scan_mutex_);
}

void CollisionManagerAlgNode::imu_callback(const sensor_msgs::Imu::ConstPtr& msg)
{
  // ROS_INFO("CollisionManagerAlgNode::imu_callback: New Message Received");

  //use appropiate mutex to shared variables if necessary
  //this->alg_.lock();
  this->imu_mutex_enter();
  this->new_imu_input_ = true;
  this->imu_frame_ = msg->header.frame_id;
  this->imu_local_acc_ = Eigen::Vector3d(msg->linear_acceleration.x, msg->linear_acceleration.y, msg->linear_acceleration.z);
  this->imu_local_ang_vel_ = Eigen::Vector3d(msg->angular_velocity.x, msg->angular_velocity.y, msg->angular_velocity.z);
  this->imu_stamp_ = msg->header.stamp;
  // if (!this->in_collision_)
  //   this->collision_start_stamp_ = msg->header.stamp;
  // if (this->in_collision_)
  //   this->collision_end_stamp_ = msg->header.stamp;
  //std::cout << msg->data << std::endl;
  //unlock previously blocked shared variables
  //this->alg_.unlock();
  this->imu_mutex_exit();
}

void CollisionManagerAlgNode::imu_mutex_enter(void)
{
  pthread_mutex_lock(&this->imu_mutex_);
}

void CollisionManagerAlgNode::imu_mutex_exit(void)
{
  pthread_mutex_unlock(&this->imu_mutex_);
}


/*  [service callbacks] */

/*  [action callbacks] */

/*  [action requests] */

void CollisionManagerAlgNode::node_config_update(Config &config, uint32_t level)
{
  this->alg_.lock();
  this->imu_mutex_enter();
  this->rear_laser_scan_mutex_enter();
  this->front_laser_scan_mutex_enter();

  if(config.rate!=this->getRate())
    this->setRate(config.rate);
  this->config_=config;

  this->front_laser_scan_mutex_exit();
  this->rear_laser_scan_mutex_exit();
  this->imu_mutex_exit();
  this->alg_.unlock();
}

void CollisionManagerAlgNode::addNodeDiagnostics(void)
{
}

/* main function */
int main(int argc,char *argv[])
{
  return algorithm_base::main<CollisionManagerAlgNode>(argc, argv, "collision_manager_alg_node");
}
